import functools
import time
from collections import deque

from django.conf import settings
from django.db import connections
from sqlalchemy import create_engine, event, util
from sqlalchemy.engine import base
from sqlalchemy.pool import NullPool
from sqlalchemy.pool import _ConnectionRecord as _ConnectionRecordBase

from .sqlite import SqliteWrapper
from .wrapper import Wrapper

__all__ = ["get_engine"]


class Cache:
    """Module level cache"""

    engines = {}


SQLALCHEMY_ENGINES = {
    "sqlite3": "sqlite",
    "mysql": "mysql",
    "postgresql": "postgresql",
    "postgresql_psycopg2": "postgresql+psycopg2",
    "oracle": "oracle",
}
SQLALCHEMY_ENGINES.update(getattr(settings, "ALDJEMY_ENGINES", {}))


def get_engine_string(alias):
    sett = connections[alias].settings_dict
    return sett["ENGINE"].rsplit(".")[-1]


def get_connection_string(alias="default"):
    engine = SQLALCHEMY_ENGINES[get_engine_string(alias)]
    options = "?charset=utf8" if engine == "mysql" else ""
    return engine + "://" + options


def get_engine(alias="default", **kwargs):
    if alias not in Cache.engines:
        engine_string = get_engine_string(alias)
        # we have to use autocommit=True, because SQLAlchemy
        # is not aware of Django transactions
        if engine_string == "sqlite3":
            kwargs["native_datetime"] = True

        pool = DjangoPool(alias=alias, creator=None)
        Cache.engines[alias] = create_engine(
            get_connection_string(alias), pool=pool, **kwargs
        )
    return Cache.engines[alias]


class DjangoPool(NullPool):
    def __init__(self, alias, *args, **kwargs):
        super(DjangoPool, self).__init__(*args, **kwargs)
        self.alias = alias
        self._aldjemy_handlers_init = False

    def status(self):
        return "DjangoPool"

    def _create_connection(self):
        return _ConnectionRecord(self, self.alias)

    def recreate(self):
        self.logger.info("Pool recreating")

        return DjangoPool(
            self.alias,
            self._creator,
            recycle=self._recycle,
            echo=self.echo,
            logging_name=self._orig_logging_name,
            use_threadlocal=self._use_threadlocal,
        )


def first_connect(engine, dbapi_connection, connection_record):
    """
    Like `sqlalchemy.engine.<locals>.first_connect`
    Without rolling back the transaction.
    https://github.com/sqlalchemy/sqlalchemy/blob/c2cad1f97c51c8a2a6ad5d371ece7bcd9c7ffcf9/lib/sqlalchemy/engine/create.py#L662
    """
    c = base.Connection(
        engine,
        connection=dbapi_connection,
        _has_events=False,
        # reconnecting will be a reentrant condition, so if the
        # connection goes away, Connection is then closed
        _allow_revalidate=False,
    )
    c._execution_options = util.EMPTY_DICT

    engine.dialect.initialize(c)


class _ConnectionRecord(_ConnectionRecordBase):
    def __init__(self, pool, alias):
        self.__pool = pool
        self.info = {}
        self.finalize_callback = deque()
        self.starttime = time.time()

        self.alias = alias
        self.wrap = False
        # Replace sqlalchemy's handler with ours
        if not pool._aldjemy_handlers_init:
            # we assume it's the last one
            previous_handler_wrapper = pool.dispatch.connect.listeners.pop()
            assert (
                previous_handler_wrapper.__name__ == "go"
            )  # wrapped by _once_unless_exception=True
            handler = previous_handler_wrapper.__closure__[0].cell_contents
            assert handler.__name__ == "first_connect", (handler, handler.__name__)

            engine = get_engine(self.alias)
            event.listen(
                pool,
                "connect",
                functools.partial(first_connect, engine),
                _once_unless_exception=True,
            )
            pool._aldjemy_handlers_init = True
        pool.dispatch.connect(self.connection, self)
        self.wrap = True

    @property
    def connection(self):
        connection = connections[self.alias]
        if connection.connection is None:
            connection._cursor()
        if connection.vendor == "sqlite":
            return SqliteWrapper(connection.connection)
        if self.wrap:
            return Wrapper(connection.connection)
        return connection.connection

    def close(self):
        pass

    def invalidate(self, e=None, soft=False):
        pass

    def get_connection(self):
        return self.connection



from django.apps import apps
from django.conf import settings
from django.db import connections, router
from django.db.backends import signals
from django.db.models.fields.related import ForeignKey, ManyToManyField, OneToOneField
from sqlalchemy import orm
from sqlalchemy.orm import registry

from .core import get_engine
from .table import generate_tables


def get_session(alias="default", recreate=False, **kwargs):
    connection = connections[alias]
    if not hasattr(connection, "sa_session") or recreate:
        engine = get_engine(alias, **kwargs)
        session = orm.sessionmaker(bind=engine)
        connection.sa_session = session()
    return connection.sa_session


def new_session(sender, connection, **kwargs):
    if connection.alias in settings.DATABASES:
        get_session(alias=connection.alias, recreate=True)


signals.connection_created.connect(new_session)


def _extract_model_attrs(metadata, model, sa_models):
    tables = metadata.tables

    name = model._meta.db_table
    qualname = (metadata.schema + "." + name) if metadata.schema else name
    table = tables[qualname]
    fks = [t for t in model._meta.fields if isinstance(t, (ForeignKey, OneToOneField))]
    attrs = {}
    rel_fields = fks + list(model._meta.many_to_many)

    for f in model._meta.fields:
        if not isinstance(f, (ForeignKey, OneToOneField)):
            if f.model != model or f.column not in table.c:
                continue  # Fields from parent model are not supported
            attrs[f.name] = orm.column_property(table.c[f.column])

    for fk in rel_fields:
        if fk.column not in table.c and not isinstance(fk, ManyToManyField):
            continue

        parent_model = fk.remote_field.model

        parent_model_meta = parent_model._meta

        if parent_model_meta.proxy:
            continue

        p_table_name = parent_model_meta.db_table
        p_table_qualname = (
            metadata.schema + "." + p_table_name if metadata.schema else p_table_name
        )
        p_table = tables[p_table_qualname]
        p_name = parent_model_meta.pk.column

        disable_backref = (
            fk.remote_field.related_name and fk.remote_field.related_name.endswith("+")
        )
        backref = (
            fk.remote_field.related_name.lower().strip("+")
            if fk.remote_field.related_name
            else None
        )
        if not backref and not disable_backref:
            backref = model._meta.object_name.lower()
            if not isinstance(fk, OneToOneField):
                backref = backref + "_set"
        elif backref and isinstance(fk, OneToOneField):
            backref = orm.backref(backref, uselist=False)

        kwargs = {}
        if isinstance(fk, ManyToManyField):
            model_pk = model._meta.pk.column
            sec_table_name = fk.remote_field.field.m2m_db_table()
            sec_table_qualname = (
                metadata.schema + "." + sec_table_name
                if metadata.schema
                else sec_table_name
            )
            sec_table = tables[sec_table_qualname]
            sec_column = fk.m2m_column_name()
            p_sec_column = fk.m2m_reverse_name()
            overlaps = [
                fk.m2m_field_name(),
                fk.m2m_reverse_field_name(),
                fk.remote_field.field.get_attname(),
                fk.remote_field.through._meta.get_field(
                    fk.remote_field.field.m2m_field_name()
                ).remote_field.related_name,
                fk.remote_field.through._meta.get_field(
                    fk.m2m_reverse_field_name()
                ).remote_field.related_name,
            ]
            kwargs.update(
                secondary=sec_table,
                primaryjoin=(sec_table.c[sec_column] == table.c[model_pk]),
                secondaryjoin=(sec_table.c[p_sec_column] == p_table.c[p_name]),
                overlaps=",".join(overlaps),
            )
            if fk.model() != model:
                backref = None
        else:
            kwargs.update(
                foreign_keys=[table.c[fk.column]],
                primaryjoin=(table.c[fk.column] == p_table.c[p_name]),
                remote_side=p_table.c[p_name],
            )
            if backref and not disable_backref:
                kwargs.update(backref=backref)
        attrs[fk.name] = orm.relationship(sa_models[parent_model], **kwargs)
    return attrs


def construct_models(metadata):
    if not metadata.tables:
        generate_tables(metadata)
    tables = metadata.tables
    models = [
        model
        for model in apps.get_models(include_auto_created=True)
        if not model._meta.proxy
    ]

    sa_models = {}

    for model in models:

        table_name = (
            metadata.schema + "." + model._meta.db_table
            if metadata.schema
            else model._meta.db_table
        )
        mixin = getattr(model, "aldjemy_mixin", None)
        bases = (mixin, BaseSQLAModel) if mixin else (BaseSQLAModel,)
        table = tables[table_name]

        # because querying happens on sqlalchemy side, we can use only one
        # type of queries for alias, so we use 'read' type
        sa_model = type(
            model._meta.object_name + ".sa",
            bases,
            {
                "table": table,
                "alias": router.db_for_read(model),
                "__module__": model.__module__,
            },
        )

        sa_models[model] = sa_model

    mapper_registry = registry()
    for model in models:
        sa_model = sa_models[model]
        table_name = (
            metadata.schema + "." + model._meta.db_table
            if metadata.schema
            else model._meta.db_table
        )
        table = tables[table_name]
        attrs = _extract_model_attrs(metadata, model, sa_models)
        mapper_registry.map_imperatively(sa_model, local_table=table, properties=attrs)

    return sa_models


class BaseSQLAModel:
    @classmethod
    def query(cls, *args, **kwargs):
        alias = getattr(cls, "alias", "default")
        if args or kwargs:
            return get_session(alias).query(*args, **kwargs)
        return get_session(alias).query(cls)



class Wrapper:
    "Wrapper to disable commit in sqla"

    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, attr):
        if attr in ["commit", "rollback"]:
            return lambda *args, **kwargs: None
        obj = getattr(self.obj, attr)
        if attr not in ["cursor", "execute"]:
            return obj
        if attr == "cursor":
            return type(self)(obj)
        return self.wrapper(obj)

    def wrapper(self, obj):
        "Implement if you need to make your customized wrapper"
        return obj

    def __call__(self, *args, **kwargs):
        self.obj = self.obj(*args, **kwargs)
        return self



from django.apps import apps
from django.conf import settings
from sqlalchemy import Column, ForeignKey, Table, types
from sqlalchemy.dialects.postgresql import ARRAY, DATERANGE, JSONB, UUID


def simple(typ):
    return lambda field: typ()


def varchar(field):
    return types.String(length=field.max_length)


def foreign_key(field):
    parent_model = field.related_model
    target = parent_model._meta
    target_table = target.db_table
    target_pk = target.pk.column
    return types.Integer, ForeignKey("%s.%s" % (target_table, target_pk))


def array_type(field):
    """
    Allows conversion of Django ArrayField to SQLAlchemy Array.
    Takes care of mapping the type of the array element.
    """
    internal_type = field.base_field.get_internal_type()

    # currently no support for multi-dimensional arrays
    if internal_type in DATA_TYPES and internal_type != "ArrayField":
        sub_type = DATA_TYPES[internal_type](field)
    else:
        raise RuntimeError("Unsupported array element type")

    return ARRAY(sub_type)


DATA_TYPES = {
    "AutoField": simple(types.Integer),
    "BigAutoField": simple(types.BigInteger),
    "BooleanField": simple(types.Boolean),
    "CharField": varchar,
    "CommaSeparatedIntegerField": varchar,
    "DateField": simple(types.Date),
    "DateTimeField": simple(types.DateTime),
    "DecimalField": lambda field: types.Numeric(
        scale=field.decimal_places, precision=field.max_digits
    ),
    "DurationField": simple(types.Interval),
    "FileField": varchar,
    "FilePathField": varchar,
    "FloatField": simple(types.Float),
    "IntegerField": simple(types.Integer),
    "BigIntegerField": simple(types.BigInteger),
    "IPAddressField": lambda field: types.CHAR(length=15),
    "NullBooleanField": simple(types.Boolean),
    "OneToOneField": foreign_key,
    "ForeignKey": foreign_key,
    "PositiveIntegerField": simple(types.Integer),
    "PositiveSmallIntegerField": simple(types.SmallInteger),
    "SlugField": varchar,
    "SmallIntegerField": simple(types.SmallInteger),
    "TextField": simple(types.Text),
    "TimeField": simple(types.Time),
    # PostgreSQL-specific types
    "ArrayField": array_type,
    "UUIDField": simple(UUID),
    "JSONField": simple(JSONB),
    "DateRangeField": simple(DATERANGE),
}


def generate_tables(metadata):
    # Update with user specified data types
    COMBINED_DATA_TYPES = dict(DATA_TYPES)
    COMBINED_DATA_TYPES.update(getattr(settings, "ALDJEMY_DATA_TYPES", {}))

    models = apps.get_models(include_auto_created=True)
    for model in models:
        name = model._meta.db_table
        qualname = (metadata.schema + "." + name) if metadata.schema else name
        if qualname in metadata.tables or model._meta.proxy:
            continue
        columns = []
        model_fields = [
            (f, f.model if f.model != model else None)
            for f in model._meta.get_fields()
            if not f.is_relation or f.one_to_one or (f.many_to_one and f.related_model)
        ]
        private_fields = model._meta.private_fields
        for field, parent_model in model_fields:
            if field not in private_fields:
                if parent_model:
                    continue

                try:
                    internal_type = field.get_internal_type()
                except AttributeError:
                    continue

                if internal_type in COMBINED_DATA_TYPES and hasattr(field, "column"):
                    typ = COMBINED_DATA_TYPES[internal_type](field)
                    if not isinstance(typ, (list, tuple)):
                        typ = [typ]
                    columns.append(
                        Column(field.column, *typ, primary_key=field.primary_key)
                    )

        Table(name, metadata, *columns)



from django.apps import AppConfig
from sqlalchemy import MetaData

from .orm import construct_models


class AldjemyConfig(AppConfig):
    name = "aldjemy"
    verbose_name = "Aldjemy"

    def ready(self):
        # Patch models with SQLAlchemy models
        for model, sa_model in construct_models(MetaData()).items():
            model.sa = sa_model



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aldjemy']

package_data = \
{'': ['*']}

install_requires = \
['Django>=2.2', 'SQLAlchemy>=1.4']

setup_kwargs = {
    'name': 'aldjemy',
    'version': '2.5',
    'description': 'SQLAlchemy for your Django models',
    'long_description': '=======\nAldjemy\n=======\n\n.. image:: https://raw.githubusercontent.com/aldjemy/aldjemy/main/logo.png\n   :alt: Aldjemy Logo\n\n|pypi_version| |pypi_license| |ci-tests|\n\n\nAldjemy integrates SQLAlchemy into an existing Django project,\nto help you build complex queries that are difficult for the Django ORM.\n\nWhile other libraries use SQLAlchemy reflection to generate SQLAlchemy models,\nAldjemy generates the SQLAlchemy models by introspecting the Django models.\nThis allows you to better control what properties in a table are being accessed.\n\n\nInstallation\n------------\n\nAdd ``aldjemy`` to your ``INSTALLED_APPS``.\nAldjemy will automatically add an ``sa`` attribute to all models,\nwhich is an SQLAlchemy ``Model``.\n\nExample:\n\n.. code-block:: python\n\n    User.sa.query().filter(User.sa.username==\'Brubeck\')\n    User.sa.query().join(User.sa.groups).filter(Group.sa.name=="GROUP_NAME")\n\nExplicit joins are part of the SQLAlchemy philosophy,\nso don\'t expect Aldjemy to be a Django ORM drop-in replacement.\nInstead, you should use Aldjemy to help with special situations.\n\n\nSettings\n--------\n\nYou can add your own field types to map django types to sqlalchemy ones with\n``ALDJEMY_DATA_TYPES`` settings parameter.\nParameter must be a ``dict``, key is result of ``field.get_internal_type()``,\nvalue must be a one arg function. You can get idea from ``aldjemy.table``.\n\nAlso it is possible to extend/override list of supported SQLALCHEMY engines\nusing ``ALDJEMY_ENGINES`` settings parameter.\nParameter should be a ``dict``, key is substring after last dot from\nDjango database engine setting (e.g. ``sqlite3`` from ``django.db.backends.sqlite3``),\nvalue is SQLAlchemy driver which will be used for connection (e.g. ``sqlite``, ``sqlite+pysqlite``).\nIt could be helpful if you want to use ``django-postgrespool``.\n\n\nMixins\n------\n\nOften django models have helper function and properties that helps to\nrepresent the model\'s data (`__str__`), or represent some model based logic.\n\nTo integrate it with aldjemy models you can put these methods into a separate mixin:\n\n.. code-block:: python\n\n    class TaskMixin:\n        def __str__(self):\n            return self.code\n\n    class Task(TaskMixin, models.Model):\n        aldjemy_mixin = TaskMixin\n        code = models.CharField(_(\'code\'), max_length=32, unique=True)\n\nVoilà! You can use ``__str__`` on aldjemy classes, because this mixin will be\nmixed into generated aldjemy model.\n\nIf you want to expose all methods and properties without creating a\nseparate mixin class, you can use the `aldjemy.meta.AldjemyMeta`\nmetaclass:\n\n.. code-block:: python\n\n    class Task(models.Model, metaclass=AldjemyMeta):\n        code = models.CharField(_(\'code\'), max_length=32, unique=True)\n\n        def __str__(self):\n            return self.code\n\nThe result is same as with the example above, only you didn\'t need to\ncreate the mixin class at all.\n\nRelease Process\n\n---------------\n\n 1. Make a Pull Request with updated changelog and bumped version of the project\n\n    .. code-block:: bash\n\n       poetry version (major|minor|patch) # choose which version to bump\n\n 2. Once the pull request is merged, create a github release with the same version, on the web console or with github cli.\n\n    .. code-block:: bash\n\n       gh release create\n\n 3. Enjoy!\n\n.. |pypi_version| image:: https://img.shields.io/pypi/v/aldjemy.svg?style=flat-square\n    :target: https://pypi.python.org/pypi/aldjemy\n    :alt: Downloads\n\n.. |pypi_license| image:: https://img.shields.io/pypi/l/aldjemy.svg?style=flat-square\n    :target: https://pypi.python.org/pypi/aldjemy\n    :alt: License\n\n.. |ci-tests| image:: https://github.com/aldjemy/aldjemy/actions/workflows/build.yml/badge.svg\n   :target: https://github.com/aldjemy/aldjemy/actions/workflows/build.yml\n   :alt: Continuous Integration results\n',
    'author': 'Mikhail Krivushin',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aldjemy/aldjemy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)
